/***************************************************************************
 *            IpFilterManager.cc
 *
 *  Copyright  2008  Fernando Tarín Morales
 *  icemanf@gmail.com
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#include "Lince/IpFilterManager.h"
#include "Lince/Utils.h"
#include "Lince/Compose.h"

#include "../config.h"
#include "Lince/i18n.h"


#include <ctype.h>

using namespace Lince;
using namespace libtorrent;

IpFilterManager::IpFilterManager()
 : ip_filter()
{
	fs::path configFile;
	configFile = build_filename(get_base_directory(), "ipfilter.dat");
	
	if (fs::exists(configFile))
		read_ipv4_file(configFile);
}

void IpFilterManager::read_ipv4_file(const fs::path &fileName)
{
	fs::ifstream in(fileName);
	while (in.good())
	{
		char line[300];
		in.getline(line, 300);
		int len = in.gcount();
		
		if (len <= 0) continue;
		if (line[0] == '#') continue;
		if (line[0] == '\n') continue;
		if (!isdigit(line[0])) continue;
		
		unsigned int a, b, c, d, flags;
		char dummy;
		
		std::stringstream ln(line);
		ln >> a >> dummy >> b >> dummy >> c >> dummy >> d >> dummy;
		address_v4 start((a << 24) + (b << 16) + (c << 8) + d);

		ln >> a >> dummy >> b >> dummy >> c >> dummy >> d >> dummy;
		address_v4 last((a << 24) + (b << 16) + (c << 8) + d);		
		
		ln >> flags;		
		
		if (flags <= 127) flags = ip_filter::blocked;
		else flags = 0;
		
		add_ipv4_rule(start, last, flags);
	}
}

void IpFilterManager::save_ipv4_file(const fs::path &fileName)
{
	fs::ofstream out(fileName, std::ios_base::binary);
	ips_v4 ips;
		
	ips = get_ips_v4();
		
	for(int i=0;i < ips.size(); i++)
	{
		if (ips[i].flags == ip_filter::blocked){
			out << ips[i].first.to_string() << "-" <<\
				ips[i].last.to_string() << ",001" << std::endl;
		}
		else{
			out << ips[i].first.to_string() << "-" <<\
				ips[i].last.to_string() << ",128" << std::endl;
		}
	}
		
	out.close();
}

ips_v4 IpFilterManager::get_filtered_ips_v4() const
{
	ips_v4 filtered_ips, temp_ips;
	
	temp_ips = get_ips_v4();
	
	for (int i=0; i < temp_ips.size(); i++)
	{
		if (temp_ips[i].flags == ip_filter::blocked)
		{
			filtered_ips.push_back(temp_ips[i]);
		}
	}
	
	return filtered_ips;
}

// Ip range should be like 127.0.0.1:127.0.0.255
void IpFilterManager::add_ipv4_rule(const std::string & range, int flags)
{
	unsigned int a, b, c, d;
	char dummy;
		
	std::stringstream ln(range);
	ln >> a >> dummy >> b >> dummy >> c >> dummy >> d >> dummy;
	libtorrent::address_v4 start((a << 24) + (b << 16) + (c << 8) + d);

	ln >> a >> dummy >> b >> dummy >> c >> dummy >> d;
	libtorrent::address_v4 last((a << 24) + (b << 16) + (c << 8) + d);
		
	if (flags <= 127) flags = ip_filter::blocked;
	else flags = 0;
		
	add_ipv4_rule(start, last, flags);
}

void IpFilterManager::add_ipv4_rule(const libtorrent::address_v4 &s, const libtorrent::address_v4 &l, int flags)
{
	std::string translated;
	
	add_rule(s, l, flags);
	
	if (flags == ip_filter::blocked)
	{
		if (s == l)
		{
			translated = String::compose(_("IP: %1 banned"), s.to_string());
			m_signal_ip_banned(translated);
		}
		else
		{
			translated = String::compose(_("IP: %1-%2 banned"), s.to_string(), l.to_string());
			m_signal_ip_banned(translated);
		}
	}
	else
	{
		if (s == l)
		{
			translated = String::compose(_("IP: %1 unbanned"), s.to_string());
			m_signal_ip_unbanned(translated);
		}
		else
		{
			translated = String::compose(_("IP: %1-%2 unbanned"), s.to_string(), l.to_string());
			m_signal_ip_unbanned(translated);
		}
	}
};

bool IpFilterManager::load_from_file_v4(const std::string &fName)
{
	fs::path fileName(fName);
	
	if (!fs::exists(fileName)) return false;
	read_ipv4_file(fileName);
	
	return true;
}

bool IpFilterManager::load_from_file_v4(const fs::path &fileName)
{
	if (!fs::exists(fileName)) return false;
	read_ipv4_file(fileName);
	
	return true;
}

bool IpFilterManager::save_to_file_v4(const std::string &fName)
{
	fs::path fileName(fName);
	save_ipv4_file(fileName);
	
	return true;
}

bool IpFilterManager::save_to_file_v4(const fs::path &fileName)
{
	save_ipv4_file(fileName);
	return true;
}

IpFilterManager::~IpFilterManager()
{
	save_ipv4_file(build_filename(get_base_directory(), "ipfilter.dat"));
}
