/***************************************************************************
 *            SetupManager.h
 *
 *  Copyright  2008  Fernando Tarín Morales
 *  icemanf@gmail.com
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#ifndef SETUPMANAGER_H
#define SETUPMANAGER_H

#include <string>
#include <cstdlib>

#include "Lince/Utils.h"

#include <boost/smart_ptr.hpp>
#include "libtorrent/intrusive_ptr_base.hpp"

#include "XMLWriter.h"
#include "XMLReader.h"

namespace Lince
{

class SetupManager;
typedef boost::intrusive_ptr<SetupManager> SetupManagerPtr;

class SetupManager: public libtorrent::intrusive_ptr_base<SetupManager>
{

public:

	static SetupManagerPtr create();

	enum StrSetting { STR_FIRST,
		SERVER_PROXY_NAME = STR_FIRST, SERVER_PROXY_USERNAME, SERVER_PROXY_PASSWORD,
		CLIENT_PROXY_NAME, CLIENT_PROXY_USERNAME, CLIENT_PROXY_PASSWORD,
		BIND_TO_INTERFACE, DEFAULT_FILE_DIRECTORY, LAST_SAVE_DIRECTORY,
		STR_LAST };

	enum IntSetting { INT_FIRST = STR_LAST + 1,
		FIRST_PORT = INT_FIRST, LAST_PORT, DHT_SUPPORT, UPNP_SUPPORT,
		LOCAL_DISCOVERY_SUPPORT, METADATA_SUPPORT, PEX_SUPPORT,
		MAX_HALF_OPEN_CONNECTIONS, SERVER_PROXY_PORT, SERVER_USE_PROXY,
		SERVER_PROXY_TYPE, CLIENT_PROXY_PORT, CLIENT_USE_PROXY, CLIENT_PROXY_TYPE,
		UPLOAD_SPEED, DISABLE_UPLOAD_SPEED, DOWNLOAD_SPEED,
		DISABLE_DOWNLOAD_SPEED, UPLOAD_SLOTS, DISABLE_UPLOAD_SLOTS,
		CONNECTIONS_NUMBER, DISABLE_CONNECTIONS_NUMBER, SHARE_RATIO_1,
		SHARE_RATIO_2, DISABLE_SHARE_RATIO, CONNECTIONS_FROM_THE_SAME_IP,
		PRIORITIZE_FIRST_PART_OF_FILES, STORAGE_TYPE, USE_ENCRYPTION,
		ENCRYPTED_INCOMING_CONNECTIONS, ENCRYPTED_OUTGOING_CONNECTIONS,
		ENCRYPTION_LEVEL, ENCRYPTION_PREFER_RC4, ACTIVE_DOWNLOADS_VALUE,
		MAX_DOWNLOADS_VALUE, ACTIVE_SHARES_VALUE, USE_IP_FILTER,
		USE_DEFAULT_FILE_DIRECTORY, USE_SCHEDULER, SCHEDULER_MONDAY,
		SCHEDULER_TUESDAY, SCHEDULER_WEDNESDAY, SCHEDULER_THURSDAY,
		SCHEDULER_FRIDAY, SCHEDULER_SATURDAY, SCHEDULER_SUNDAY,
		SCHEDULER_HOUR0_DOWN, SCHEDULER_HOUR0_UP, SCHEDULER_HOUR1_DOWN,
		SCHEDULER_HOUR1_UP, SCHEDULER_HOUR2_DOWN, SCHEDULER_HOUR2_UP,
		SCHEDULER_HOUR3_DOWN, SCHEDULER_HOUR3_UP, SCHEDULER_HOUR4_DOWN,
		SCHEDULER_HOUR4_UP, SCHEDULER_HOUR5_DOWN, SCHEDULER_HOUR5_UP,
		SCHEDULER_HOUR6_DOWN, SCHEDULER_HOUR6_UP, SCHEDULER_HOUR7_DOWN,
		SCHEDULER_HOUR7_UP, SCHEDULER_HOUR8_DOWN, SCHEDULER_HOUR8_UP,
		SCHEDULER_HOUR9_DOWN, SCHEDULER_HOUR9_UP, SCHEDULER_HOUR10_DOWN,
		SCHEDULER_HOUR10_UP, SCHEDULER_HOUR11_DOWN, SCHEDULER_HOUR11_UP,
		SCHEDULER_HOUR12_DOWN, SCHEDULER_HOUR12_UP, SCHEDULER_HOUR13_DOWN,
		SCHEDULER_HOUR13_UP, SCHEDULER_HOUR14_DOWN, SCHEDULER_HOUR14_UP,
		SCHEDULER_HOUR15_DOWN, SCHEDULER_HOUR15_UP, SCHEDULER_HOUR16_DOWN,
		SCHEDULER_HOUR16_UP, SCHEDULER_HOUR17_DOWN, SCHEDULER_HOUR17_UP,
		SCHEDULER_HOUR18_DOWN, SCHEDULER_HOUR18_UP, SCHEDULER_HOUR19_DOWN,
		SCHEDULER_HOUR19_UP, SCHEDULER_HOUR20_DOWN, SCHEDULER_HOUR20_UP,
		SCHEDULER_HOUR21_DOWN, SCHEDULER_HOUR21_UP, SCHEDULER_HOUR22_DOWN,
		SCHEDULER_HOUR22_UP, SCHEDULER_HOUR23_DOWN, SCHEDULER_HOUR23_UP,
		SMARTBAN_SUPPORT,
		INT_LAST };

	enum Int64Setting { INT64_FIRST = INT_LAST + 1,
		TOTAL_UPLOAD = INT64_FIRST, TOTAL_DOWNLOAD, TOTAL_TIME, INT64_LAST,
		SETTINGS_LAST = INT64_LAST };


	const std::string& get(StrSetting key, bool useDefault = true) const
	{
		return (isSet[key] || !useDefault)
			? strSettings[key - STR_FIRST] : strDefaults[key - STR_FIRST];
	}

	int get(IntSetting key, bool useDefault = true) const
	{
		return (isSet[key] || !useDefault)
			? intSettings[key - INT_FIRST] : intDefaults[key - INT_FIRST];
	}

	bool getBool(IntSetting key, bool useDefault = true) const
	{
		return (get(key, useDefault) != 0);
	}

	int64_t get(Int64Setting key, bool useDefault = true) const
	{
		return (isSet[key] || !useDefault)
			? int64Settings[key - INT64_FIRST] : int64Defaults[key - INT64_FIRST];
	}

	void set(StrSetting key, std::string const& value)
	{
		if(((key == SERVER_PROXY_USERNAME) || (key == SERVER_PROXY_PASSWORD))
		   && (value.size() > 35))
		{
			strSettings[key - STR_FIRST] = value.substr(0, 35);
		}
		else
		{
			strSettings[key - STR_FIRST] = value;
		}
		isSet[key] = !value.empty();
	}

	void set(IntSetting key, int value)
	{
		intSettings[key - INT_FIRST] = value;
		isSet[key] = true;
	}

	void set(IntSetting key, const std::string& value)
	{
		if(value.empty())
		{
			intSettings[key - INT_FIRST] = 0;
			isSet[key] = false;
		}
		else
		{
			intSettings[key - INT_FIRST] = to_int(value.c_str());
			isSet[key] = true;
		}
	}

	void set(Int64Setting key, int64_t value)
	{
		int64Settings[key - INT64_FIRST] = value;
		isSet[key] = true;
	}

	void set(Int64Setting key, int value)
	{
		int64Settings[key - INT64_FIRST] = (int64_t) value;
		isSet[key] = true;
	}

	void set(Int64Setting key, const std::string& value)
	{
		if(value.empty())
		{
			int64Settings[key - INT64_FIRST] = 0;
			isSet[key] = false;
		}
		else
		{
			int64Settings[key - INT64_FIRST] = to_int64(value.c_str());
			isSet[key] = true;
		}
	}

	void set(IntSetting key, bool value) { set(key, (int)value); }

	void setDefault(StrSetting key, std::string const& value)
	{
		strDefaults[key - STR_FIRST] = value;
	}

	void setDefault(IntSetting key, int value)
	{
		intDefaults[key - INT_FIRST] = value;
	}

	void setDefault(Int64Setting key, int64_t value)
	{
		int64Defaults[key - INT64_FIRST] = value;
	}

	void load();
	void save();

	bool isDefault(int aSet) { return !isSet[aSet]; }

	void setFileName(const std::string config_file_path)
	{
		config_file = config_file_path;
	};

	~SetupManager();

private:

	SetupManager();

	std::string config_file;

	XMLReader reader;
	XMLWriter writer;

	static const std::string settingTags[SETTINGS_LAST+1];

    std::string strSettings[STR_LAST - STR_FIRST];
	int intSettings[INT_LAST - INT_FIRST];
	int64_t int64Settings[INT64_LAST - INT64_FIRST];

	std::string strDefaults[STR_LAST - STR_FIRST];
	int intDefaults[INT_LAST - INT_FIRST];
	int64_t int64Defaults[INT64_LAST - INT64_FIRST];

	bool isSet[SETTINGS_LAST];
};

} //Namespace

#endif
