/***************************************************************************
 *            TorrentsManager.h
 *
 *  Copyright  2008  Fernando Tarín Morales
 *  icemanf@gmail.com
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#ifndef TORRENTSMANAGER_H
#define TORRENTSMANAGER_H

#include <deque>
#include <list>
#include <map>

#include <boost/signal.hpp>
#include <boost/bind.hpp>

#include "libtorrent/session.hpp"
#include "libtorrent/torrent_handle.hpp"
#include "libtorrent/torrent_info.hpp"
#include "libtorrent/intrusive_ptr_base.hpp"

#include "Torrent.h"

namespace Lince
{

class TorrentsManager;

typedef boost::intrusive_ptr<TorrentsManager> TorrentsManagerPtr;
typedef std::deque<Torrent> torrentList;
typedef std::pair<int, bool> torrentPosition;
	
typedef boost::signal<void (int)> intSignal;
	
class TorrentsManager: public libtorrent::intrusive_ptr_base<TorrentsManager>
{
	torrentList downloadingTorrents, sharedTorrents;
	std::list<int> stoppedDownloadingTorrents, queuedDownloadingTorrents, activeDownloadingTorrents;
	std::list<int> stoppedSharedTorrents, queuedSharedTorrents, activeSharedTorrents;
	
	TorrentsManager();
	
	torrentList::iterator get_downloading_torrent_iterator_at(int position);
	torrentList::iterator get_shared_torrent_iterator_at(int position);
	
	void start_all_downloading_torrents();
	void stop_all_downloading_torrents();
	void start_all_shared_torrents();
	void stop_all_shared_torrents();
	
	typedef std::map<libtorrent::sha1_hash, torrentPosition> torrentPositions;
	torrentPositions tPositions;
	
	intSignal m_signal_torrent_switched;
	
	void on_finished_torrent(int position);
	void on_stopped_torrent(const torrentPosition &);
	void on_started_torrent(const torrentPosition &);
	
public:
	// Funciones miembro nuevas usando torrentPosition.
	torrentPosition get_torrent_position(const libtorrent::sha1_hash &);
		
	Torrent* get_torrent_at(const torrentPosition &);
		
	void start_all_torrents();
	void stop_all_torrents();
		
	boost::signals::connection switched_torrent_signal(intSignal::slot_function_type subscriber)
	{
		return m_signal_torrent_switched.connect(subscriber);
	};

	int torrents_count(){ 
		return (downloadingTorrents.size() + sharedTorrents.size()); 
	};
		
	// Downloading torrents list management
	void append_downloading_torrent(Torrent&);
	void remove_downloading_torrent_at(int position);
		
	Torrent* get_downloading_torrent_at(int position);
	torrentList::iterator get_downloading_torrent_iterator_at_begin();
	torrentList::iterator get_downloading_torrent_iterator_at_end();
		
	void move_downloading_torrent(int, int);
	
	void start_downloading_torrent_at(int);
	void stop_downloading_torrent_at(int);
			
	int downloading_torrents_count()
	{
		return downloadingTorrents.size();
	};
		
	void rebuild_downloading_torrents_queue();
		
	// Shared torrents list management
	void append_shared_torrent(Torrent&);
	void remove_shared_torrent_at(int position);
	
	Torrent* get_shared_torrent_at(int position);		
	torrentList::iterator get_shared_torrent_iterator_at_begin();
	torrentList::iterator get_shared_torrent_iterator_at_end();
	
	void move_shared_torrent(int, int);
	
	void start_shared_torrent_at(int);
	void stop_shared_torrent_at(int);
		
	int shared_torrents_count()
	{
		return sharedTorrents.size();
	};
		
	void rebuild_shared_torrents_queue();		
		
	static TorrentsManagerPtr create();
	virtual ~TorrentsManager();
};
	
} // Namespace

#endif
