/***************************************************************************
 *            SessionManager.cc
 *
 *  Copyright  2008  Fernando Tarín Morales
 *  icemanf@gmail.com
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#include "Lince/SessionManager.h"

#include "Lince/Core.h"
#include "Lince/SetupManager.h"
#include "Lince/Utils.h"
#include "Lince/XMLReader.h"
#include "Lince/XMLWriter.h"

#include "libtorrent/file_pool.hpp"
#include "libtorrent/magnet_uri.hpp"
#include "libtorrent/extensions/ut_pex.hpp"
#include <libtorrent/extensions/ut_metadata.hpp>
#include "libtorrent/extensions/metadata_transfer.hpp"
#include <libtorrent/extensions/smart_ban.hpp>

#include "../config.h"

#include <boost/filesystem.hpp>

using namespace Lince;
using namespace libtorrent;

SessionManagerPtr SessionManager::create()
{
	return SessionManagerPtr(new SessionManager());
}

SessionManager::SessionManager()
 : session(libtorrent::fingerprint("LC", 1, 0, 0, 0), 0)
{
	active_shared_torrents = 0;
	currentDownloadingSize = 0;
	totalDownloadingSize = totalSharedSize = 0;

	// TODO: Funcion obsoleta
	set_severity_level(alert::info);
	ip_filter = IpFilterManagerPtr(new IpFilterManager());

	// Update settings
	update_settings();

	// Load torrents of the previous session. If any.
	load_resume_data();

	// Launch session updater thread
	update_thread = boost::thread(boost::bind(&SessionManager::session_update_manager
											  , this));

	Core::get_alerts_manager()->
		metadata_received_signal(boost::bind(&SessionManager::on_metadata_received,
					this, _1));
}

void SessionManager::check_alerts(const boost::system::error_code& /*e*/,
    boost::asio::deadline_timer* t)
{
    Core::get_alerts_manager()->check_alerts();

    t->expires_at(t->expires_at() + boost::posix_time::seconds(1));
    t->async_wait(boost::bind(&SessionManager::check_alerts, this,
    	boost::asio::placeholders::error, t));
}

void SessionManager::update_sizes(const boost::system::error_code& /*e*/,
	    boost::asio::deadline_timer* t)
{
	TorrentsManagerPtr tm = Core::get_torrents_manager();
	torrentList::iterator i;

	currentDownloadingSize = totalDownloadingSize = 0;

	for (i=tm->get_downloading_torrent_iterator_at_begin();
			i!=tm->get_downloading_torrent_iterator_at_end(); ++i)
	{
		currentDownloadingSize += i->get_downloaded();
		totalDownloadingSize += i->get_size();
	}

	totalSharedSize = 0;

	for (i=tm->get_shared_torrent_iterator_at_begin();
			i!=tm->get_shared_torrent_iterator_at_end(); ++i)
	{
		totalSharedSize += i->get_size();
	}

	t->expires_at(t->expires_at() + boost::posix_time::seconds(1));
	t->async_wait(boost::bind(&SessionManager::update_sizes, this,
	    	boost::asio::placeholders::error, t));
}

void SessionManager::check_scheduler(const boost::system::error_code& /*e*/,
    boost::asio::deadline_timer* t)
{
	update_upload_session_speed();
	update_download_session_speed();

	t->expires_at(t->expires_at() + boost::posix_time::hours(1));
	t->async_wait(boost::bind(&SessionManager::check_scheduler, this,
		  boost::asio::placeholders::error, t));
}

void SessionManager::session_update_manager()
{
	boost::asio::deadline_timer alerts_timer(io, boost::posix_time::seconds(1));
	boost::asio::deadline_timer sizes_timer(io, boost::posix_time::seconds(1));
	boost::posix_time::ptime now = boost::posix_time::second_clock::local_time();
	int minutes = 60 - now.time_of_day().minutes();
	int seconds = (minutes * 60) - now.time_of_day().seconds() + 5;
	boost::asio::deadline_timer scheduler_timer(io,
												boost::posix_time::seconds(seconds));

	alerts_timer.async_wait(boost::bind(&SessionManager::check_alerts, this,
										boost::asio::placeholders::error,
										&alerts_timer));

	sizes_timer.async_wait(boost::bind(&SessionManager::update_sizes, this,
										boost::asio::placeholders::error,
										&sizes_timer));

	scheduler_timer.async_wait(boost::bind(&SessionManager::check_scheduler, this,
										   boost::asio::placeholders::error,
										   &scheduler_timer));
	io.run();
}

void SessionManager::update_ports()
{
	SetupManagerPtr s = Core::get_setup_manager();

	try{
		if (s->get(SetupManager::BIND_TO_INTERFACE) != "")
		{
			listen_on(std::make_pair(s->get(SetupManager::FIRST_PORT),
					s->get(SetupManager::LAST_PORT)),
					s->get(SetupManager::BIND_TO_INTERFACE).c_str());
		}
		else
			listen_on(std::make_pair(s->get(SetupManager::FIRST_PORT),
					s->get(SetupManager::LAST_PORT)));
	}
	catch(std::exception& e){
		listen_on(std::make_pair(s->get(SetupManager::FIRST_PORT),
					s->get(SetupManager::LAST_PORT)));
	}
}

// TODO: Check that casting works fine.
void SessionManager::update_encryption_support()
{
	SetupManagerPtr s = Core::get_setup_manager();
	pe_settings encryption_settings;

	if (s->get(SetupManager::USE_ENCRYPTION))
	{
		encryption_settings.out_enc_policy = (pe_settings::enc_policy)
			s->get(SetupManager::ENCRYPTED_OUTGOING_CONNECTIONS);

		encryption_settings.in_enc_policy = (pe_settings::enc_policy)
			s->get(SetupManager::ENCRYPTED_INCOMING_CONNECTIONS);

		encryption_settings.allowed_enc_level = (pe_settings::enc_level)
			s->get(SetupManager::ENCRYPTION_LEVEL);

		encryption_settings.prefer_rc4 =
			s->getBool(SetupManager::ENCRYPTION_PREFER_RC4);
	}
	else{
		encryption_settings.in_enc_policy = pe_settings::disabled;
		encryption_settings.out_enc_policy = pe_settings::disabled;
	}

	set_pe_settings(encryption_settings);
}

// TODO: Set proxy type.
void SessionManager::update_proxy_settings()
{
	SetupManagerPtr s = Core::get_setup_manager();
	proxy_settings proxy_data;

	if (s->get(SetupManager::SERVER_USE_PROXY))
	{
		proxy_data.hostname = s->get(SetupManager::SERVER_PROXY_NAME);
		proxy_data.port = s->get(SetupManager::SERVER_PROXY_PORT);
		proxy_data.username = s->get(SetupManager::SERVER_PROXY_USERNAME);
		proxy_data.password = s->get(SetupManager::SERVER_PROXY_PASSWORD);
	}
	else
	{
		proxy_data.type = proxy_settings::none;
	}
	set_tracker_proxy(proxy_data);
	set_web_seed_proxy(proxy_data);

	if (s->get(SetupManager::CLIENT_USE_PROXY))
	{
		proxy_data.hostname = s->get(SetupManager::CLIENT_PROXY_NAME);
		proxy_data.port = s->get(SetupManager::CLIENT_PROXY_PORT);
		proxy_data.username = s->get(SetupManager::CLIENT_PROXY_USERNAME);
		proxy_data.password = s->get(SetupManager::CLIENT_PROXY_PASSWORD);
	}
	else
	{
		proxy_data.type = proxy_settings::none;
	}
	set_peer_proxy(proxy_data);
	set_dht_proxy(proxy_data);
}

void SessionManager::update_upnp()
{
	SetupManagerPtr s = Core::get_setup_manager();
	if (s->get(SetupManager::UPNP_SUPPORT))
	{
		// UPnP port mapping
		start_upnp();
		// NAT-PMP port mapping
		start_natpmp();
	}
	else
	{
		stop_upnp();
		stop_natpmp();
	}
}

// Local service discovery (finds peers on the local network)
void SessionManager::update_lsd()
{
	SetupManagerPtr s = Core::get_setup_manager();
	if (s->get(SetupManager::LOCAL_DISCOVERY_SUPPORT))
		start_lsd();
	else
		stop_lsd();
}

void SessionManager::update_dht()
{
	SetupManagerPtr s = Core::get_setup_manager();
	if (s->getBool(SetupManager::DHT_SUPPORT))
	{
		boost::filesystem::ifstream dht_state_file(build_filename(get_base_directory(),\
			std::string("dhtState")), std::ios_base::binary);

		dht_state_file.unsetf(std::ios_base::skipws);
		entry dht_state;

		try{
			dht_state = bdecode(std::istream_iterator<char>(dht_state_file)\
								, std::istream_iterator<char>());
		}
		catch (std::exception&) {}

		start_dht(dht_state);
		add_dht_router(std::make_pair(std::string("router.bittorrent.com"), 6881));
		add_dht_router(std::make_pair(std::string("router.utorrent.com"), 6881));
		add_dht_router(std::make_pair(std::string("router.bitcomet.com"), 6881));
	}
	else
	{
		stop_dht();
	}
}

void SessionManager::update_settings()
{
	SetupManagerPtr s = Core::get_setup_manager();
	session_settings settings;

	// User agent
	settings.user_agent = PACKAGE_STRING;
	settings.use_dht_as_fallback = false;

	// Config Options
	update_upnp();
	update_lsd();

	// Port Options
	update_ports();

	// Encryption Support
	update_encryption_support();

	// Proxy options.
	update_proxy_settings();

	// DHT
	update_dht();

	// Ip Filter
	if (s->getBool(SetupManager::USE_IP_FILTER)){
		update_ip_filter();
	}

	// Upload speed
	update_upload_session_speed();

	// Download speed
	update_download_session_speed();

	// Max number of connections
	if(!s->getBool(SetupManager::DISABLE_CONNECTIONS_NUMBER))
	{
		set_max_connections(s->get(SetupManager::CONNECTIONS_NUMBER));
	}

	// Max number of uploads
	if (!s->getBool(SetupManager::DISABLE_UPLOAD_SLOTS))
	{
		set_max_uploads(s->get(SetupManager::UPLOAD_SLOTS));
	}

	// Max half open connections
	if (s->get(SetupManager::MAX_HALF_OPEN_CONNECTIONS))
	{
		set_max_half_open_connections(
				s->get(SetupManager::MAX_HALF_OPEN_CONNECTIONS));
	}
	else
	{
		set_max_half_open_connections(-1);
	}

	// Support for metadata.
	if (s->getBool(SetupManager::METADATA_SUPPORT))
	{
		add_extension(&create_metadata_plugin);
		add_extension(&create_ut_metadata_plugin);
	}

	// Support for pex.
	if (s->getBool(SetupManager::PEX_SUPPORT))
	{
		add_extension(&create_ut_pex_plugin);
	}

	// Support for smart ban.
	if (s->getBool(SetupManager::SMARTBAN_SUPPORT))
	{
		add_extension(&libtorrent::create_smart_ban_plugin);
	}

	// Allow multiple connections from the same ip
	settings.allow_multiple_connections_per_ip =
		s->getBool(SetupManager::CONNECTIONS_FROM_THE_SAME_IP);

	set_settings(settings);
}

void SessionManager::update_ip_filter()
{
	SetupManagerPtr s = Core::get_setup_manager();

	if (s->getBool(SetupManager::USE_IP_FILTER))
		set_ip_filter(*ip_filter);
	else
		set_ip_filter(libtorrent::ip_filter());
}

void SessionManager::load_resume_data()
{
	int number=1, number_files, share_ratio1, share_ratio2;
	std::string status;
	XMLReader reader;
	torrent_localinfo tl;
	std::ostringstream torrent_number;
	boost::filesystem::path torrent_path;

	reader.setFileName(build_filename_as_string(get_base_directory(),\
					   std::string("torrentpaths.xml")));

	torrent_number << "Torrent" << number;

	while(reader.findChild(torrent_number.str()))
	{
		reader.goToNextElement();
		tl.file_name = reader.getChildData();

		reader.goToNextElement();
		tl.save_path = reader.getChildData();

		reader.goToNextElement();
		status = reader.getChildData();

		// TODO: Revisar el allocate
		reader.goToNextElement();
		tl.storage_mode = to_int(reader.getChildData());

		reader.goToNextElement();
		tl.is_complete = to_int(reader.getChildData());

		reader.goToNextElement();
		tl.total_uploaded = to_int64(reader.getChildData());

		reader.goToNextElement();
		tl.max_upload = to_int(reader.getChildData());

		reader.goToNextElement();
		tl.max_download = to_int(reader.getChildData());

		reader.goToNextElement();
		tl.max_connects = to_int(reader.getChildData());

		reader.goToNextElement();
		tl.max_upload_slots = to_int(reader.getChildData());

		reader.goToNextElement();
		share_ratio1 = to_int(reader.getChildData());

		reader.goToNextElement();
		share_ratio2 = to_int(reader.getChildData());

		tl.share_ratio  = static_cast<float>(share_ratio1 +
						   static_cast<float>(share_ratio2 / 10.0));

		reader.goToNextElement();
		tl.total_time = to_int64(reader.getChildData());

		reader.goToNextElement();
		tl.failed_hashes = to_int(reader.getChildData());

		reader.goToNextElement();
		number_files = to_int(reader.getChildData());

		std::vector<int> priority_files;
		reader.goToNextElement();
		if (!reader.isEmpty())
		{
			for (int i=0; i < number_files; i++)
			{
				reader.goToNextElement();
				priority_files.push_back(to_int(reader.getChildData()));
			}
			tl.file_priorities = priority_files;
		}

		reader.goToNextElement();
		if (!reader.isEmpty())
		{
			std::vector<std::string> names_vector;
			for (int i=0; i < number_files; i++)
			{
				std::string name;

				reader.goToNextElement();
				reader.goToNextElement();
				names_vector.push_back(reader.getChildData());
			}
			tl.file_new_names = names_vector;
		}

		reader.goToNextElement();
		reader.goToNextElement();
		std::vector<libtorrent::announce_entry> trackers;
		int number_trackers = to_int(reader.getChildData());

		for (int i=0; i < number_trackers; i++)
		{
			reader.goToNextElement();
			libtorrent::announce_entry tracker(reader.getChildData());
			trackers.push_back(tracker);
		}

		tl.trackers = trackers;

		// TODO: Que hacer con esto?? Seguramente sacarlo y meterlo en la configuracion del gui.
		//reader.goToNextElement();
		/*showTab = Utils::toInt(reader.getChildData());*/

		number++;
		torrent_number.str("");
		torrent_number << "Torrent" << number;

		if (status == "Stopped")
			tl.is_stopped = true;
		else
			tl.is_stopped = false;

		add_torrent_to_session(tl);
	}
}


void SessionManager::save_torrents_local_data(){
	int number = 1;
	std::string paths_filename;
	std::ostringstream torrent_number;
	XMLWriter writer;
	TorrentsManagerPtr tm = Core::get_torrents_manager();
	torrentList::iterator iter_begin, iter_end, i;

	paths_filename = build_filename_as_string(get_base_directory(),
											  std::string("torrentpaths.xml"));

	writer.setFileName(paths_filename);
	writer.startDocument("Lince");
	writer.startElement("Paths");

	for (int j=0; j<2; j++)
	{
		if (j)
		{
			iter_begin = tm->get_downloading_torrent_iterator_at_begin();
			iter_end = tm->get_downloading_torrent_iterator_at_end();
		}
		else
		{
			iter_begin = tm->get_shared_torrent_iterator_at_begin();
			iter_end = tm->get_shared_torrent_iterator_at_end();
		}

		for (i = iter_begin; i != iter_end; ++i)
		{
			torrent_number << "Torrent" << number;
			writer.startElement(torrent_number.str());

			writer.startElement("Name");
			writer.addAttribute("type", "string");
			writer.addElementData(i->file_name());
			writer.endElement();

			writer.startElement("Path");
			writer.addAttribute("type", "string");
			writer.addElementData(i->save_path().native_directory_string());
			writer.endElement();

			writer.startElement("Status");
			writer.addAttribute("type", "string");
			if (i->is_stopped())
				writer.addElementData("Stopped");
			else
				writer.addElementData("Active");
			writer.endElement();

			writer.startElement("StorageMode");
			writer.addAttribute("type", "int");
			writer.addElementData(static_cast<int>(i->storage_mode()));
			writer.endElement();

			writer.startElement("Complete");
			writer.addAttribute("type", "int");
			writer.addElementData(i->is_complete());
			writer.endElement();

			writer.startElement("Uploaded");
			writer.addAttribute("type", "int64");
			writer.addElementData(i->total_uploaded());
			writer.endElement();

			writer.startElement("MaxUploadSpeed");
			writer.addAttribute("type", "int");
			writer.addElementData(i->max_upload_speed());
			writer.endElement();

			writer.startElement("MaxDownloadSpeed");
			writer.addAttribute("type", "int");
			writer.addElementData(i->max_download_speed());
			writer.endElement();

			writer.startElement("MaxConnections");
			writer.addAttribute("type", "int");
			writer.addElementData(i->max_connects());
			writer.endElement();

			writer.startElement("MaxUploadSlots");
			writer.addAttribute("type", "int");
			writer.addElementData(i->upload_slots());
			writer.endElement();

			writer.startElement("ShareRatioValue1");
			writer.addAttribute("type", "int");
			writer.addElementData(static_cast<int>(i->share_ratio()));
			writer.endElement();

			writer.startElement("ShareRatioValue2");
			writer.addAttribute("type", "int");
			writer.addElementData(static_cast<int>(static_cast<float>((i->share_ratio() - static_cast<int>(i->share_ratio())) * 10)));
			writer.endElement();

			writer.startElement("Time");
			writer.addAttribute("type", "int64");
			writer.addElementData(i->elapsed_time());
			writer.endElement();

			writer.startElement("FailedHashes");
			writer.addAttribute("type", "int");
			writer.addElementData(i->failed_hashes());
			writer.endElement();

			writer.startElement("TorrentFiles");
			writer.addAttribute("type", "int");
			writer.addElementData(i->get_torrent_info().num_files());
			writer.endElement();

			writer.startElement("PriorityFiles");
			std::vector<int> fp = i->file_priorities();
			for (int j=0; j < fp.size(); ++j)
			{
				std::ostringstream file_number;
				file_number << "File" << (j+1);
				writer.startElement(file_number.str());
				writer.addAttribute("type", "int");
				writer.addElementData(fp[j]);
				writer.endElement();
				file_number.str("");
			}
			writer.endElement();

			writer.startElement("NewNames");
			std::vector<std::string> t = i->file_new_names();
			for (int j=0; j<t.size(); j++)
			{
				std::ostringstream file_number;
				file_number << "File" << (j+1);
				writer.startElement(file_number.str());
				writer.startElement("Name");
				writer.addAttribute("type", "string");
				writer.addElementData(t[j]);
				writer.endElement();
				writer.endElement();
			}
			writer.endElement();

			writer.startElement("Trackers");
			std::vector<announce_entry> trackers = i->trackers();
			writer.startElement("NumberOfTrackers");
			writer.addAttribute("type", "int");
			writer.addElementData((int) trackers.size());
			writer.endElement();
			for (int j=0; j<trackers.size(); j++)
			{
				writer.startElement("Tracker");
				writer.addAttribute("type", "string");
				writer.addElementData(trackers[j].url);
				writer.endElement();
			}
			writer.endElement();

			writer.endElement();
			torrent_number.str("");
			number++;
		}
	}

	writer.endDocument();
}

void SessionManager::add_torrent_to_session(Lince::torrent_localinfo &tl)
{
	SetupManagerPtr s = Core::get_setup_manager();
	boost::filesystem::path torrent_path = build_filename(get_torrents_base_directory(),\
								 tl.file_name);
	bool pauseTorrent;

	boost::intrusive_ptr<torrent_info> t;
	try
	{
		t = new torrent_info(torrent_path);
	}
	catch (std::exception&)
	{
		return;
	}

	if (!tl.file_new_names.empty()){
		libtorrent::file_storage &fs = t->files();

		for (int i=0; i < tl.file_new_names.size(); i++)
			fs.rename_file(i, tl.file_new_names[i]);
	}
	torrent_handle th;
	add_torrent_params p;
	lazy_entry resume_data;
	std::ostringstream hash_str;

	hash_str << t->info_hash();

	std::vector<char> buf;
	if (load_file(build_filename(get_resume_data_base_directory(), hash_str.str())\
			, buf) == 0)
		p.resume_data = &buf;

	p.ti = t;
	p.save_path = tl.save_path;
	p.storage_mode = (storage_mode_t) tl.storage_mode;
	p.paused = false;
	p.duplicate_is_error = true;
	p.auto_managed = false;
	th = add_torrent(p);

	if (!tl.file_priorities.empty())
	{
		th.prioritize_files(tl.file_priorities);
	}

	th.resolve_countries(true);

	th.set_max_connections(tl.max_connects);
	th.set_max_uploads(tl.max_upload_slots);
	th.set_ratio(tl.share_ratio);
	th.replace_trackers(tl.trackers);

	Torrent torrent(th);

	// Set stored data
	torrent.set_complete(tl.is_complete);
	torrent.set_max_upload_speed(tl.max_upload);
	torrent.set_max_download_speed(tl.max_download);
	torrent.set_elapsed_time(tl.total_time);
	torrent.set_failed_hashes(tl.failed_hashes);
	torrent.set_total_uploaded(tl.total_uploaded);
	torrent.set_share_ratio(tl.share_ratio);
	torrent.set_storage_mode(p.storage_mode);
	torrent.set_upload_slots(tl.max_upload_slots);
	torrent.set_max_connects(tl.max_connects);
	torrent.set_file_priorities(tl.file_priorities);
	torrent.set_file_new_names(tl.file_new_names);
	torrent.set_file_name(tl.file_name);
	if (tl.is_stopped)
		torrent.stop();

	if (s->get(SetupManager::PRIORITIZE_FIRST_PART_OF_FILES))
		torrent.prioritize_first_part_of_files(2, 10485760);

	Core::get_alerts_manager()->add_local_alert
		(local_alert(TORRENT_ADDED_PREVIOUS_SESSION, WARNING, torrent.name(), torrent.info_hash()));

	TorrentsManagerPtr tm = Core::get_torrents_manager();
	if (tl.is_complete)
		tm->append_shared_torrent(torrent);
	else
		tm->append_downloading_torrent(torrent);
}

void SessionManager::on_metadata_received(libtorrent::torrent_handle &th)
{
	SetupManagerPtr s = Core::get_setup_manager();
	TorrentsManagerPtr tm = Core::get_torrents_manager();

	// If the queue limit has been reached just return
	if (s->getBool(SetupManager::MAX_DOWNLOADS_VALUE) &&
		(tm->downloading_torrents_count() == s->get(SetupManager::MAX_DOWNLOADS_VALUE)))
		return;

	float share_ratio;
	if (!s->get(SetupManager::DISABLE_SHARE_RATIO)){
		share_ratio = static_cast<float>(s->get(SetupManager::SHARE_RATIO_1) +
					static_cast<float>(s->get(SetupManager::SHARE_RATIO_2) / 10.0));
	}
	else{
		share_ratio =  0.0;
	}

	th.set_ratio(share_ratio);
	th.resolve_countries(true);

	std::vector<int> torrent_priorities;
	for (int i = 0; i < th.get_torrent_info().num_files(); i++)
	{
		torrent_priorities.push_back(1);
	}

	Torrent torrent(th);

	torrent.set_share_ratio(share_ratio);
	torrent.set_file_name(th.name());
	torrent.set_file_priorities(torrent_priorities);

	// Store the torrent for future sessions
	std::string filename = Lince::build_filename_as_string(
			Lince::get_torrents_base_directory().native_file_string(), th.name());
	boost::filesystem::basic_ofstream<char> out;
	Lince::save_entry(filename, torrent.create_torrent());

	if (s->get(SetupManager::PRIORITIZE_FIRST_PART_OF_FILES))
		torrent.prioritize_first_part_of_files(2, 10485760);

	Core::get_alerts_manager()->add_local_alert
		(local_alert(TORRENT_METADATA_ADDED, WARNING, torrent.name(), torrent.info_hash()));

	tm->append_downloading_torrent(torrent);
	save_torrents_local_data();
}

void SessionManager::add_torrent_to_session(const boost::filesystem::path & torrent_path,
	const std::string & save_path, const std::vector<int> & priorityfiles,
	std::vector<std::string> &torrentfilenewnames, bool stop)
{
	boost::intrusive_ptr<torrent_info> t;
	try
	{
		t = new torrent_info(torrent_path);
	}
	catch (std::exception&)
	{
		return;
	}

	SetupManagerPtr s = Core::get_setup_manager();
	TorrentsManagerPtr tm = Core::get_torrents_manager();
	torrent_handle th;
	add_torrent_params p;
	int storage = s->get(SetupManager::STORAGE_TYPE);

	if (s->getBool(SetupManager::MAX_DOWNLOADS_VALUE) &&
		(tm->downloading_torrents_count() == s->get(SetupManager::MAX_DOWNLOADS_VALUE)))
		throw queue_limit_reached();

	if (!torrentfilenewnames.empty()){
		libtorrent::file_storage &fs = t->files();

		for (int i=0; i < torrentfilenewnames.size(); i++)
			fs.rename_file(i, torrentfilenewnames[i]);
	}

	p.ti = t;
	p.save_path = save_path;
	if (storage == 1)
		p.storage_mode = libtorrent::storage_mode_allocate;
	else if (storage == 2)
		p.storage_mode = libtorrent::storage_mode_compact;
	else if (storage == 3)
		p.storage_mode = libtorrent::storage_mode_sparse;
	p.paused = false;
	p.duplicate_is_error = true;
	p.auto_managed = false;

	th = add_torrent(p);

	// Copy torrent
	if (!exists(boost::filesystem::path(build_filename(get_torrents_base_directory(), torrent_path.leaf())))){
		boost::filesystem::copy_file(torrent_path, boost::filesystem::path(build_filename(get_torrents_base_directory(), torrent_path.leaf())));
	}

	if (!priorityfiles.empty())
		th.prioritize_files(priorityfiles);

	float share_ratio;
	if (!s->get(SetupManager::DISABLE_SHARE_RATIO)){
		share_ratio = static_cast<float>(s->get(SetupManager::SHARE_RATIO_1) +
					static_cast<float>(s->get(SetupManager::SHARE_RATIO_2) / 10.0));
	}
	else{
		share_ratio =  0.0;
	}

	th.set_ratio(share_ratio);
	th.resolve_countries(true);

	Torrent torrent(th);

	torrent.set_share_ratio(share_ratio);
	torrent.set_file_name(torrent_path.leaf());
	torrent.set_file_priorities(priorityfiles);
	torrent.set_file_new_names(torrentfilenewnames);

	if (stop)
		torrent.stop();

	if (s->get(SetupManager::PRIORITIZE_FIRST_PART_OF_FILES))
		torrent.prioritize_first_part_of_files(2, 10485760);

	Core::get_alerts_manager()->add_local_alert
		(local_alert(TORRENT_ADDED, WARNING, torrent.name(), torrent.info_hash()));

	tm->append_downloading_torrent(torrent);
	save_torrents_local_data();
}

void SessionManager::add_magnet_to_session(const std::string &magnet,
		const std::string &save_path)
{
	SetupManagerPtr s = Core::get_setup_manager();
	TorrentsManagerPtr tm = Core::get_torrents_manager();

	if (s->getBool(SetupManager::MAX_DOWNLOADS_VALUE) &&
		(tm->downloading_torrents_count() == s->get(SetupManager::MAX_DOWNLOADS_VALUE)))
		throw queue_limit_reached();

	if (!s->getBool(SetupManager::METADATA_SUPPORT))
		throw metadata_not_activated();

	// TODO: This retrieves the sha1 of the magnet link but should be improved
	int pos = magnet.find_first_of('&');
	std::string sha1 = magnet.substr(8, pos-8);
	pos = magnet.find_last_of(':');
	sha1 = sha1.substr(pos+1, sha1.length() - pos);

	add_torrent_params p;
	int storage = s->get(SetupManager::STORAGE_TYPE);

	p.info_hash = sha1_hash(sha1);
	p.save_path = save_path;
	if (storage == 1)
		p.storage_mode = libtorrent::storage_mode_allocate;
	else if (storage == 2)
		p.storage_mode = libtorrent::storage_mode_compact;
	else if (storage == 3)
		p.storage_mode = libtorrent::storage_mode_sparse;
	p.paused = false;
	p.duplicate_is_error = true;
	p.auto_managed = false;

	libtorrent::torrent_handle th =libtorrent::add_magnet_uri(*this, magnet, p);
	Core::get_alerts_manager()->add_local_alert(
			local_alert(MAGNET_LINK_ADDED, WARNING, th.name(), th.info_hash()));
}

void SessionManager::save_torrents_resume_data()
{
	TorrentsManagerPtr tm = Core::get_torrents_manager();
	torrentList::iterator iter;

	pause();

	// Downloading Torrents
	for (iter = tm->get_downloading_torrent_iterator_at_begin();
		 iter != tm->get_downloading_torrent_iterator_at_end();
		 ++iter)
		iter->save_resume_data();

	// Shared Torrents
	for (iter = tm->get_shared_torrent_iterator_at_begin();
		 iter != tm->get_shared_torrent_iterator_at_end();
		 ++iter)
		iter->save_resume_data();
}


void SessionManager::remove_downloading_torrent(int position, bool remove_data)
{
	TorrentsManagerPtr tm = Core::get_torrents_manager();
	SetupManagerPtr sm = Core::get_setup_manager();
	boost::filesystem::path local_torrent_path, resume_data_path;
	std::string file_name = tm->get_downloading_torrent_at(position)->file_name();
	libtorrent::torrent_handle th = tm->get_downloading_torrent_at(position)->get_handle();

	tm->remove_downloading_torrent_at(position);

	local_torrent_path = build_filename(get_torrents_base_directory(),
										file_name);
	resume_data_path = build_filename(get_resume_data_base_directory(),
									  file_name);

	Core::get_alerts_manager()->add_local_alert
		(local_alert(TORRENT_REMOVED, WARNING, th.name(),
					 th.get_torrent_info().info_hash()));

	if (remove_data){
		remove_torrent(th, libtorrent::session::delete_files);
	}
	else{
		remove_torrent(th);
	}

	boost::filesystem::remove(local_torrent_path);
	boost::filesystem::remove(resume_data_path);

	save_torrents_local_data();
}

void SessionManager::remove_shared_torrent(int position)
{
	TorrentsManagerPtr tm = Core::get_torrents_manager();
	SetupManagerPtr sm = Core::get_setup_manager();
	boost::filesystem::path local_torrent_path, resume_data_path;
	std::string file_name = tm->get_shared_torrent_at(position)->file_name();
	libtorrent::torrent_handle th = tm->get_shared_torrent_at(position)->get_handle();

	tm->remove_shared_torrent_at(position);

	local_torrent_path = build_filename(get_torrents_base_directory(),
										file_name);
	resume_data_path = build_filename(get_resume_data_base_directory(),
									  file_name);

	Core::get_alerts_manager()->add_local_alert
		(local_alert(TORRENT_REMOVED, WARNING, th.name(),
					 th.get_torrent_info().info_hash()));

	remove_torrent(th);

	boost::filesystem::remove(local_torrent_path);
	boost::filesystem::remove(resume_data_path);

	save_torrents_local_data();
}


void SessionManager::recheck_torrent(const torrentPosition &tp)
{
	Torrent *t;

	t = Core::get_torrents_manager()->get_torrent_at(tp);

	if (t->get_status() == libtorrent::torrent_status::checking_files ||
		t->get_status() == libtorrent::torrent_status::queued_for_checking)
			return;

	t->force_recheck();

	/*entry e;
	load_entry(build_filename(get_torrents_base_directory(), t->file_name()), e);

	boost::intrusive_ptr<libtorrent::torrent_info>
		info(new libtorrent::torrent_info(e));
	boost::filesystem::path save_path = t->save_path();
	libtorrent::entry empty;
	storage_mode_t storage_mode = t->storage_mode();

	remove_torrent(t->get_handle());
	torrent_handle th = add_torrent(info, save_path, empty, storage_mode);
	t->set_handle(th);
	*/
}

void SessionManager::set_upload_speed_session(int speed)
{
	SetupManagerPtr s = Core::get_setup_manager();
	if (!s->get(SetupManager::USE_SCHEDULER))
	{
		if (speed != -1)
			set_upload_rate_limit(speed * 1024);
		else
			set_upload_rate_limit(speed);
	}
}

void SessionManager::set_download_speed_session(int speed)
{
	SetupManagerPtr s = Core::get_setup_manager();
	if (!s->get(SetupManager::USE_SCHEDULER))
	{
		if (speed != -1)
			set_download_rate_limit(speed * 1024);
		else
			set_download_rate_limit(speed);
	}
}

void SessionManager::update_upload_session_speed()
{
	int speed=-1;
	SetupManagerPtr s = Core::get_setup_manager();
	if (s->get(SetupManager::USE_SCHEDULER))
	{
		bool current_day = false;
		short day = boost::gregorian::day_clock::local_day().day_of_week();
		switch (day){
			case 0:
				if (s->get(SetupManager::SCHEDULER_SUNDAY)) current_day = true;
				break;
			case 1:
				if (s->get(SetupManager::SCHEDULER_MONDAY)) current_day = true;
				break;
			case 2:
				if (s->get(SetupManager::SCHEDULER_TUESDAY)) current_day = true;
				break;
			case 3:
				if (s->get(SetupManager::SCHEDULER_WEDNESDAY)) current_day = true;
				break;
			case 4:
				if (s->get(SetupManager::SCHEDULER_THURSDAY)) current_day = true;
				break;
			case 5:
				if (s->get(SetupManager::SCHEDULER_FRIDAY)) current_day = true;
				break;
			case 6:
				if (s->get(SetupManager::SCHEDULER_SATURDAY)) current_day = true;
				break;
		}

		if (!current_day)
		{
			if (!s->get(SetupManager::DISABLE_UPLOAD_SPEED))
				speed = s->get(SetupManager::UPLOAD_SPEED);
			else
				speed = 0;
		}
		else
		{
			boost::posix_time::ptime now = boost::posix_time::second_clock::local_time();
			short hour = now.time_of_day().hours();

			switch (hour)
			{
				case 0:
					speed = s->get(SetupManager::SCHEDULER_HOUR0_UP);
					break;
				case 1:
					speed = s->get(SetupManager::SCHEDULER_HOUR1_UP);
					break;
				case 2:
					speed = s->get(SetupManager::SCHEDULER_HOUR2_UP);
					break;
				case 3:
					speed = s->get(SetupManager::SCHEDULER_HOUR3_UP);
					break;
				case 4:
					speed = s->get(SetupManager::SCHEDULER_HOUR4_UP);
					break;
				case 5:
					speed = s->get(SetupManager::SCHEDULER_HOUR5_UP);
					break;
				case 6:
					speed = s->get(SetupManager::SCHEDULER_HOUR6_UP);
					break;
				case 7:
					speed = s->get(SetupManager::SCHEDULER_HOUR7_UP);
					break;
				case 8:
					speed = s->get(SetupManager::SCHEDULER_HOUR8_UP);
					break;
				case 9:
					speed = s->get(SetupManager::SCHEDULER_HOUR9_UP);
					break;
				case 10:
					speed = s->get(SetupManager::SCHEDULER_HOUR10_UP);
					break;
				case 11:
					speed = s->get(SetupManager::SCHEDULER_HOUR11_UP);
					break;
				case 12:
					speed = s->get(SetupManager::SCHEDULER_HOUR12_UP);
					break;
				case 13:
					speed = s->get(SetupManager::SCHEDULER_HOUR13_UP);
					break;
				case 14:
					speed = s->get(SetupManager::SCHEDULER_HOUR14_UP);
					break;
				case 15:
					speed = s->get(SetupManager::SCHEDULER_HOUR15_UP);
					break;
				case 16:
					speed = s->get(SetupManager::SCHEDULER_HOUR16_UP);
					break;
				case 17:
					speed = s->get(SetupManager::SCHEDULER_HOUR17_UP);
					break;
				case 18:
					speed = s->get(SetupManager::SCHEDULER_HOUR18_UP);
					break;
				case 19:
					speed = s->get(SetupManager::SCHEDULER_HOUR19_UP);
					break;
				case 20:
					speed = s->get(SetupManager::SCHEDULER_HOUR20_UP);
					break;
				case 21:
					speed = s->get(SetupManager::SCHEDULER_HOUR21_UP);
					break;
				case 22:
					speed = s->get(SetupManager::SCHEDULER_HOUR22_UP);
					break;
				case 23:
					speed = s->get(SetupManager::SCHEDULER_HOUR23_UP);
					break;
			}
		}
	}
	else
	{
		if (!s->get(SetupManager::DISABLE_UPLOAD_SPEED))
			speed = s->get(SetupManager::UPLOAD_SPEED);
		else
			speed = 0;
	}

	if (speed)
		set_upload_rate_limit(speed * 1024);
	else
		set_upload_rate_limit(-1);
}

void SessionManager::update_download_session_speed()
{
	int speed=-1;
	SetupManagerPtr s = Core::get_setup_manager();
	if (s->get(SetupManager::USE_SCHEDULER))
	{
		bool current_day = false;
		short day = boost::gregorian::day_clock::local_day().day_of_week();
		switch (day){
			case 0:
				if (s->get(SetupManager::SCHEDULER_SUNDAY)) current_day = true;
				break;
			case 1:
				if (s->get(SetupManager::SCHEDULER_MONDAY)) current_day = true;
				break;
			case 2:
				if (s->get(SetupManager::SCHEDULER_TUESDAY)) current_day = true;
				break;
			case 3:
				if (s->get(SetupManager::SCHEDULER_WEDNESDAY)) current_day = true;
				break;
			case 4:
				if (s->get(SetupManager::SCHEDULER_THURSDAY)) current_day = true;
				break;
			case 5:
				if (s->get(SetupManager::SCHEDULER_FRIDAY)) current_day = true;
				break;
			case 6:
				if (s->get(SetupManager::SCHEDULER_SATURDAY)) current_day = true;
				break;
		}

		if (!current_day)
		{
			if (!s->get(SetupManager::DISABLE_DOWNLOAD_SPEED))
				speed = s->get(SetupManager::DOWNLOAD_SPEED);
			else
				speed = 0;
		}
		else
		{
			boost::posix_time::ptime now = boost::posix_time::second_clock::local_time();
			short hour = now.time_of_day().hours();

			switch (hour)
			{
				case 0:
					speed = s->get(SetupManager::SCHEDULER_HOUR0_DOWN);
					break;
				case 1:
					speed = s->get(SetupManager::SCHEDULER_HOUR1_DOWN);
					break;
				case 2:
					speed = s->get(SetupManager::SCHEDULER_HOUR2_DOWN);
					break;
				case 3:
					speed = s->get(SetupManager::SCHEDULER_HOUR3_DOWN);
					break;
				case 4:
					speed = s->get(SetupManager::SCHEDULER_HOUR4_DOWN);
					break;
				case 5:
					speed = s->get(SetupManager::SCHEDULER_HOUR5_DOWN);
					break;
				case 6:
					speed = s->get(SetupManager::SCHEDULER_HOUR6_DOWN);
					break;
				case 7:
					speed = s->get(SetupManager::SCHEDULER_HOUR7_DOWN);
					break;
				case 8:
					speed = s->get(SetupManager::SCHEDULER_HOUR8_DOWN);
					break;
				case 9:
					speed = s->get(SetupManager::SCHEDULER_HOUR9_DOWN);
					break;
				case 10:
					speed = s->get(SetupManager::SCHEDULER_HOUR10_DOWN);
					break;
				case 11:
					speed = s->get(SetupManager::SCHEDULER_HOUR11_DOWN);
					break;
				case 12:
					speed = s->get(SetupManager::SCHEDULER_HOUR12_DOWN);
					break;
				case 13:
					speed = s->get(SetupManager::SCHEDULER_HOUR13_DOWN);
					break;
				case 14:
					speed = s->get(SetupManager::SCHEDULER_HOUR14_DOWN);
					break;
				case 15:
					speed = s->get(SetupManager::SCHEDULER_HOUR15_DOWN);
					break;
				case 16:
					speed = s->get(SetupManager::SCHEDULER_HOUR16_DOWN);
					break;
				case 17:
					speed = s->get(SetupManager::SCHEDULER_HOUR17_DOWN);
					break;
				case 18:
					speed = s->get(SetupManager::SCHEDULER_HOUR18_DOWN);
					break;
				case 19:
					speed = s->get(SetupManager::SCHEDULER_HOUR19_DOWN);
					break;
				case 20:
					speed = s->get(SetupManager::SCHEDULER_HOUR20_DOWN);
					break;
				case 21:
					speed = s->get(SetupManager::SCHEDULER_HOUR21_DOWN);
					break;
				case 22:
					speed = s->get(SetupManager::SCHEDULER_HOUR22_DOWN);
					break;
				case 23:
					speed = s->get(SetupManager::SCHEDULER_HOUR23_DOWN);
					break;
			}
		}
	}
	else
	{
		if (!s->get(SetupManager::DISABLE_DOWNLOAD_SPEED))
			speed = s->get(SetupManager::DOWNLOAD_SPEED);
		else
			speed = 0;
	}


	if (speed)
		set_download_rate_limit(speed * 1024);
	else
		set_download_rate_limit(-1);
}


SessionManager::~SessionManager()
{
	SetupManagerPtr s = Core::get_setup_manager();
	TorrentsManagerPtr tm = Core::get_torrents_manager();

	// Finish update thread
	io.stop();
	update_thread.join();

	// Update session statistics
	s->set(SetupManager::TOTAL_UPLOAD, s->get(SetupManager::TOTAL_UPLOAD) +
		   static_cast<boost::int64_t>(status().total_upload));

	s->set(SetupManager::TOTAL_DOWNLOAD, s->get(SetupManager::TOTAL_DOWNLOAD) +
		   static_cast<boost::int64_t>(status().total_download));

	s->set(SetupManager::TOTAL_TIME, s->get(SetupManager::TOTAL_TIME) +
		   static_cast<boost::int64_t>(get_session_elapsed_time()));

	save_torrents_local_data();
	save_torrents_resume_data();

	int num_resume_data = tm->torrents_count();

	// Manage resume data alerts
	while (num_resume_data > 0){
		alert const* a = wait_for_alert(libtorrent::seconds(5));
		// if we don't get an alert within 5 seconds, abort
		if (a == 0) break;

		std::auto_ptr<alert> holder = pop_alert();
		if (dynamic_cast<save_resume_data_failed_alert const*>(a)){
			--num_resume_data;
			continue;
		}

		save_resume_data_alert const* rd = dynamic_cast<save_resume_data_alert const*>(a);
		if (rd == 0)
			continue;

		std::ostringstream hash_str;

		hash_str << rd->handle.get_torrent_info().info_hash();

		save_entry(build_filename(get_resume_data_base_directory(),
								  hash_str.str()),
				   *rd->resume_data);


		--num_resume_data;
	}

	if (s->getBool(SetupManager::DHT_SUPPORT))
		save_entry(build_filename(get_base_directory(),std::string("dhtState")),\
						dht_state());
}
